from population import Population
from individual import Individual
import matplotlib.pyplot as plt
import time
import numpy as np
from metrics import metrics
import copy
import random
import math


class HMOA:
    def __init__(self,problem, popsize, max_evo):
        # pareto front
        self.pareto_f = []
        # population
        self.Pop = []
        # offspring
        self.Qt = []
        # combined population
        self.Rt = []
        # other classes
        self.populations = Population(popsize,problem)
        self.idv = Individual()
        self.popsize = popsize
        self.max_evo = max_evo
        self.problem = problem
        self.prob = 0
        self.metric = metrics('HMOA')

    def fast_nodominate_sort(self, P):
        F = []
        F1 = []
        Sp = [[] for i in range(len(P))]
        Np = [0 for i in range(len(P))]
        for j, p in enumerate(P):
            for i, q in enumerate(P):
                if j != i:
                    if self.is_dominate(p, q):
                        Sp[j].append(i)
                    elif self.is_dominate(q, p):
                        Np[j] += 1
            if Np[j] == 0:
                p.p_rank = 1
                F1.append(j)
        
        i = 0
        while len(F1) != 0:
            F.append(F1)
            Q = []
            for pi in F1:
                for q in Sp[pi]:
                    Np[q] -= 1
                    if Np[q] == 0:
                        P[q].p_rank = i + 2
                        Q.append(q)
            i = i + 1
            F1 = Q
        return F

    def is_dominate(self, a, b):
        # determine if a dominate b
        a_f = a.F_value
        b_f = b.F_value
        i = 0
        for av, bv in zip(a_f, b_f):
            if av < bv:
                i = i + 1
            if av > bv:
                return False
        if i != 0:
            return True
        return False

    def crowding_dist(self, Fi):
        # calculate crowding distance
        f_max = Fi[0].F_value[:]
        f_min = Fi[0].F_value[:]
        f_num = len(f_max)
        for p in Fi:
            p.dp = 0
        Fi_len = len(Fi)
        for m in range(f_num):
            Fi = sorted(Fi, key = lambda x: x.F_value[m])
            Fi[0].dp = float('inf')
            Fi[Fi_len - 1].dp = float('inf')
            f_min[m] = Fi[0].F_value[m]
            f_max[m] = Fi[Fi_len - 1].F_value[m]
        
            for f in range(1, Fi_len - 2):
                a = Fi[f + 1].F_value[m] - Fi[f - 1].F_value[m]
                b = f_max[m] - f_min[m]
                if b != 0:
                    Fi[f].dp = Fi[f].dp + a / b
                else:
                    Fi[f].dp = Fi[f].dp + 0


    def inv_append(self, idx, P, to):
        for i in idx:
            to.append(P[i])

    def find_pf(self, P):
        F1 = []
        for j, p in enumerate(P):
            n = 0
            for i, q in enumerate(P):
                if j != i:
                    if self.is_dominate(q, p):
                        n += 1
            if n == 0:
                F1.append(p)
        return F1  

    def update1(self, pop, cp):
        delete = []
        new_pop = []
        for index_ in range(len(pop)):
            p = pop[index_]
            if self.is_dominate(p, cp) or cp.F_value == p.F_value:
                return [False, None]
            elif self.is_dominate(cp, p):
                delete.append(index_)

        for i, p in enumerate(pop):
            if i not in delete:
                new_pop.append(p)
        new_pop.append(cp)
        return [True, new_pop]

    def PLS(self, pf):
        new_pf = []
        for p in pf:
            if p.F == True:
                rate = 0
            else:
                rate = 1
            if random.random() < rate:
                new_pf.append(p)

        operations = ['id_greedy', '2opt', 'id_random']
        
        # local search
        n = 0
        while new_pf != [] and n < 5:
            pa = []
            for p in new_pf:            
                cp = self.idv.copy_one(p)
                s_truck = cp.X[0]
                s_drone = cp.X[1]
                for operator in  operations:
                    [new_truck, new_drone, s_new] = self.problem.neighborhoods(s_truck, s_drone, operator)
                    ccp = self.idv.make_a_pop(new_truck, new_drone, s_new)
                    if self.is_dominate(ccp, p):
                        [flag, pop] = self.update1(pf, ccp)
                        if flag == True:
                            pf = pop
                            [flag, pa] = self.update1(pa, ccp)
            new_pf = pa
            n += 1

        pf1 =[]
        for p in pf:
            p.F = True
            pf1.append(p)
        return pf1


    def store_to_pf(self, pf, fi):
        # find new solutions
        solutions = [p.F_value for p in pf]
        for p in fi:
            if p.F_value not in solutions:
                pf.append(p)
        pf = self.find_pf(pf)        
        return pf


    def run(self):
        # start time
        start_time = time.time()

        self.Pop = self.populations.creat_pop(self.popsize)

        gen = 0
        while gen < self.max_evo:
            gen += 1
            # Rt：combined population
            self.Rt = []
            # offspring
            self.Qt = self.populations.next_Pop(self.Pop, rate = [0.8,0.3])
            # combine
            self.Rt = self.Pop + self.Qt
            self.Rt = self.populations.move_duplication(self.Rt, self.pareto_f)
            # find fronts
            F = self.fast_nodominate_sort(self.Rt)
            Pt_next = []
            i = 0
            # enviroment selection
            while (len(Pt_next) + len(F[i])) < self.populations.popsize:
                Fi = []
                for k in F[i]:
                    Fi.append(self.Rt[k])

                if i == 0:
                    self.pareto_f = self.store_to_pf(self.pareto_f, Fi)
                    if random.random() < self.prob:
                        self.pareto_f = self.PLS(self.pareto_f)
                                            
                    Fi = self.pareto_f
                    self.metric.record(self.pareto_f,gen)
                    print('%s th pareto len %s:' % (gen, len(self.pareto_f)))

                self.crowding_dist(Fi)
                Pt_next += Fi 
                i += 1
            Fi = []
            # add the last front
            self.inv_append(F[i], self.Rt, Fi)
            Fi = sorted(Fi, key = lambda x: x.dp, reverse= True)
            Pt_next = Pt_next + Fi
            # only select the first pop_size individual
            Pt_next = Pt_next[0:self.populations.popsize]
            self.Pop = Pt_next
            self.prob += 1/self.max_evo

        end_time = time.time()
        used_time = end_time - start_time
        print('used time:', used_time)
        self.metric.recordTime(used_time)
        self.metric.recordfinal(self.pareto_f)


    def draw(self,problemname, outpath, repeat):

        pf1_data = []
        pf2_data = []

        for p in self.pareto_f:
            pf1_data.append(p.F_value[0])
            if p.F_value[1] != 0:
                pf2_data.append(p.F_value[1])
            else:
                pf2_data.append(0)
        f1_data = []
        f2_data = []
        for a in self.Rt:
            f1_data.append(a.F_value[0])
            if a.F_value[1] != 0:
                f2_data.append(a.F_value[1])
            else:
                f2_data.append(0)
        
        plt.figure() 
        plt.xlabel('Cost', fontsize=15)
        plt.ylabel('-Customer satisfaction', fontsize=15)

        plt.scatter(f1_data, f2_data, c='black', s=5)
        plt.scatter(pf1_data, pf2_data, c='red', s=10)
        plt.tight_layout()
        #plt.show()
        #plt.savefig(outpath + problemname.split('.')[0] + '_' + str(repeat) + '.jpg', dpi = 300)


    def plot_map(self):
        color_ = ['#FF0000','#BC8F8F','#FA8072','#FAA460','#2E8B57','#FFF5EE','#A0522D','#C0C0C0','#87CEEB','#6A5ACD','#708090']
        # map
        X = []
        Y = []
        for cus in self.problem.data_dict:
            x = self.problem.data_dict[cus]['latDeg']
            y = self.problem.data_dict[cus]['lonDeg']
            X.append(x)
            Y.append(y)

        p = self.pareto_f[int(len(self.pareto_f)/2)]

        s_truck = p.X[0]
        s_drone = p.X[1]

        X_truck = []
        Y_truck = []

        for cus in s_truck:
            x = self.problem.data_dict[cus]['latDeg']
            y = self.problem.data_dict[cus]['lonDeg']
            X_truck.append(x)
            Y_truck.append(y)

        plt.figure()
        plt.scatter(X,Y)
        plt.plot(X_truck,Y_truck,color = 'blue',label = 'Truck route')

        drawed = []
        
        for index_ in range(s_drone.shape[1]):
            i_index = s_drone[0][index_]      
            k_index = s_drone[2][index_]
            i       = s_truck[i_index]
            j       = s_drone[1][index_]   
            k       = s_truck[k_index]
            d       = s_drone[3][index_]

            X2 = []
            Y2 = []

            ix = self.problem.data_dict[i]['latDeg']
            iy = self.problem.data_dict[i]['lonDeg']
            X2.append(ix)
            Y2.append(iy)
            jx = self.problem.data_dict[j]['latDeg']
            jy = self.problem.data_dict[j]['lonDeg']
            X2.append(jx)
            Y2.append(jy)
            kx = self.problem.data_dict[k]['latDeg']
            ky = self.problem.data_dict[k]['lonDeg']
            X2.append(kx)
            Y2.append(ky)

            if index_ != s_drone.shape[1] - 1:
                if d not in drawed:
                    plt.plot(X2,Y2,'--',color = color_[d], label = 'Drone {} route'.format(d+1))
                    drawed.append(d)
                else:
                    plt.plot(X2,Y2,'--',color = color_[d])

        # last drone route
        if d not in drawed:
            plt.plot(X2,Y2,'--',color = color_[d], label = 'Drone {} route'.format(d+1))
        else:
            plt.plot(X2,Y2,'--',color = color_[d])
        print('drone:',s_drone.shape[1])
        # depot = 0
        depot_x = [self.problem.data_dict[0]['latDeg']]
        depot_y = [self.problem.data_dict[0]['lonDeg']]
        plt.scatter(depot_x, depot_y, c='yellow', marker='*')
        plt.legend()
        plt.tight_layout()

        plt.show()


    def save(self, time, problemname, outpath):
        pf1_data = []
        pf2_data = []

        for p in self.pareto_f:
            pf1_data.append(str(p.F_value[0]))
            pf2_data.append(str(p.F_value[1]))

        with open(outpath + problemname.split('.')[0] + '_results.csv', 'a+') as f:
            data = ','.join(pf1_data)
            f.write(data)
            f.write('\t')
            data = ','.join(pf2_data)
            f.write(data)
            f.write('\t')
            f.write(str(time))
            f.write('\n')
        #with open(outpath + problemname.split('.')[0] + '_RunTime.csv', 'a+') as f:

    
    def draw_HV(self):
        # paretoFs = self.metric.paretoFs
        # refpoint = self.metric.define_refpoints(paretoFs)
        worest_node = np.array([999, 0])
        best_node = np.array([0, -(len(self.problem.data_dict) - 1)])
        self.metric.cal_HV(worest_node, best_node)

        X = self.metric.gen
        Y = self.metric.HVs

        X1 = []
        Y1 = []
        for i in range(0, len(X),10):
            x = X[i]
            X1.append(x)
            y = Y[i]
            Y1.append(y)

        plt.figure()
        plt.plot(X1,Y1, label = 'HV')
        plt.xlabel('Generations', fontsize=15)
        plt.ylabel('Hypervolume', fontsize=15)
        plt.legend()
        plt.tight_layout()
        plt.show()

    def draw_cost(self):
        self.metric.cal_fitness()

        X = self.metric.gen
        Y1 = self.metric.MeanCost
        Y2 = self.metric.MinCost

        plt.figure()
        plt.plot(X,Y1, label = 'MeanCost')
        plt.xlabel('Generations', fontsize=15)
        plt.ylabel('MeanCost', fontsize=15)
        plt.legend()
        plt.tight_layout()
        plt.show()

        plt.figure()
        plt.plot(X,Y2, label = 'MinCost')
        plt.xlabel('Generations', fontsize=15)
        plt.ylabel('MinCost', fontsize=15)
        plt.legend()
        plt.tight_layout()
        plt.show()