import numpy as np
import platypus as plat

class metrics():
    def __init__(self,AlgorithmName):
        self.algorithm = AlgorithmName
        #self.generation = 0
        self.paretoFs = []
        self.runTime = 0
        #self.solutions = []
        self.HVs = []
        self.gen = []
        self.MeanCost = []
        self.MinCost = []
        self.MeanSatis = []
        self.MinSatis = []
        self.finalsolution = []
        self.finalPFs = []


    def record(self, pareto_f,generations):
        front = []
        #front_solution = []
        for p in pareto_f:
            front.append([p.F_value[0], p.F_value[1]])
            #front_solution.append(p.X)
        self.paretoFs.append(front)
        #self.solutions.append(front_solution)
        self.gen.append(generations)

    def recordTime(self, time):
        self.runTime = time

    def define_refpoints(self, paretoFs):
        value1 = []
        value2 = []
        for pareto in paretoFs:
            for point in pareto:
                value1.append(point[0])
                value2.append(point[1])

        worest_value1 = sorted(value1, reverse = True)[0]
        worest_value2 = sorted(value2, reverse = True)[0]
        return [worest_value1, worest_value2]

    def normalization(self, point, worest_node, best_node):
        for m in range(len(point)):
            point[m] = abs((point[m] - best_node[m])/(worest_node[m] - best_node[m]))
        return point


    def cal_HV(self, worest_node, best_node):
        problem = plat.ZDT1() # we do not use this function, just use it to trigger the plat function...
        hyp = plat.indicators.Hypervolume(minimum=best_node, maximum=worest_node)

        paretofs = self.paretoFs
        for i in range(len(self.paretoFs)):
            pf = []
            s = []
            for j in range(len(self.paretoFs[i])):
                if paretofs[i][j] not in s:
                    s.append(paretofs[i][j])
                    solution = plat.Solution(problem)
                    solution.objectives = paretofs[i][j]
                    pf.append(solution)            
            if len(pf) != 0:
                HV = hyp.calculate(pf)
            else:
                print('warning: HV is zero')
                HV = 0
            self.HVs.append(HV)

        print('HV of the pareto front of {} is {}:'.format(self.algorithm, self.HVs[-1]))

    def cal_fitness(self):
        for paretoF in self.paretoFs:
            cost = []
            satisfaction = []
            for point in paretoF:
                cost.append(point[0])
                satisfaction.append(point[1])

            self.MeanCost.append(sum(cost)/len(cost))
            self.MinCost.append(min(cost))
            self.MeanSatis.append(sum(satisfaction)/len(satisfaction))
            self.MinSatis.append(min(satisfaction))


    def C_metric(self, paretoFront1, paretoFront2):
        n = 0
        for i in paretoFront2:
            flag = 1
            for j in paretoFront1:
                if self.is_weakdominate(j, i) == False:
                    flag = 0
            if flag == 1:
                n +=1
        c = n/len(paretoFront2)
        print('c_metric of two pareto fronts:', c)
        return c  

    def is_weakdominate(self, a_f, b_f):
        # a是否支配b
        i = 0
        for av, bv in zip(a_f, b_f):
            if av <= bv:
                i = i + 1
            if av > bv:
                return False
        if i != 0:
            return True
        return False


    def recordfinal(self, pareto_f):
        front = []
        front_solution = []
        for p in pareto_f:
            front.append([p.F_value[0], p.F_value[1]])
            front_solution.append(p.X)
        self.finalPFs.append(front)
        self.finalsolution.append(front_solution)










