from individual import Individual
import numpy as np
import random
from tools import *
import copy
import math

class Population:

    def __init__(self,popsize,problem):
        self.idv = Individual()
        self.popsize = popsize
        self.problem = problem
        self.crossover_rate = 0
        self.mutation_rate = 1
        self.refpoint   = [0, -(len(self.problem.data_dict)-1)]
        self.nadirpoint = [0, 0]

    def creat_pop(self,popsize, refpoint = None, weights = None):
        P = []
        ref_x = []
        ref_y = []
        for i in range(popsize):
            p = self.idv.creat_one()
            solution = self.problem.initialsolution()
            p.X = solution[1]
            p.F_value = solution[0]
            p.index = i
            if refpoint == True:
                ref_x.append(solution[0][0])
                ref_y.append(solution[0][1])
            P.append(p)
        
        if refpoint == True:
            self.nadirpoint  = np.array([max(ref_x), max(ref_y)])
            # self.refpoint = np.array([min(ref_x), min(ref_y)])
            for i, p in enumerate(P):
                for m in range(2):
                    p.F +=  normalization(p.F_value[m], self.nadirpoint[m], self.refpoint[m]) * weights[i][m]
        return P

    def generate_a_pop(self, s_truck, s_drone, s_value, p, weights = []):
        k  = p.index 
        cp = self.idv.copy_one(p)
        cp.X[0] = s_truck
        cp.X[1] = s_drone
        cp.F_value = s_value
        cp.F = 0
        # for MOEAD, weights are considered
        if weights != []:
            for m in range(2):
                cp.F += normalization(cp.F_value[m], self.nadirpoint[m], self.refpoint[m]) * weights[k][m]
        return cp

    def update_refpoint(self, pop):
        ref_x = []
        ref_y = []        
        for p in pop:
            ref_x.append(p.F_value[0])
            ref_y.append(p.F_value[1])
        self.refpoint  = np.array([max(ref_x), max(ref_y)])
        self.nadirpoint = np.array([min(ref_x), min(ref_y)])

    def binary_tournament(self, population):
        parent_a = random.choice(population)
        parent_b = random.choice(population)
        return self.crowded_comparison(parent_a, parent_b)

    def crowded_comparison(self,parent_a,parent_b):
        if parent_a.p_rank < parent_b.p_rank:
            return parent_a
        elif parent_b.p_rank < parent_a.p_rank:
            return parent_b
        elif parent_a.dp > parent_b.dp:
            return parent_a
        else:
            return parent_b

    def next_Pop(self, Pop, rate = [0.8, 0.3]):
        self.crossover_rate = rate[0]
        self.mutation_rate  = rate[1]
        # crossover operation is useless and makes solution infeasible
        child_P = []
        while len(child_P) < self.popsize:
            parent_a = self.binary_tournament(Pop)
            child = self.idv.copy_one(parent_a)
            if random.random() < self.crossover_rate:
                parent_b = self.binary_tournament(Pop)
                children = self.crossover(parent_a, parent_b)
            else:
                children = [child]
            for child_to_muation in children:
                if len(child_P) < self.popsize:
                    if random.random() < self.mutation_rate:
                        p = self.multimode_mutation(child_to_muation)
                        child_P.append(p)
                    else:
                        child_P.append(child_to_muation)
        return child_P                    

    def multimode_mutation(self, p):
        cp = self.idv.copy_one(p)
        s_truck = cp.X[0]
        s_drone = cp.X[1]
        if len(s_truck) <= self.problem.LTL:
            operation = random.choice(['d2t', 'swap'])
        elif s_drone.shape[1] <= self.problem.drone_num:
            operation = random.choice(['t2d', 'swap'])
        else:
            operation = random.choice(['t2d', 'swap', 'd2t'])

        [new_truck, new_drone, s] = self.problem.neighborhoods(s_truck, s_drone, operation)

        cp.X[0] = new_truck
        cp.X[1] = new_drone
        cp.F_value = s

        return cp   

    def reproduction(self,p1, p2, mutation_rate = 0.3):
        child_P =[]
        p1 = self.idv.reset_one(p1)
        p2 = self.idv.reset_one(p2)
        children = self.crossover(p1, p2)
        for child_to_muation in children:
            if random.random() < mutation_rate:
                p = self.multimode_mutation(child_to_muation)
                child_P.append(p)
            else:
                child_P.append(child_to_muation)
        return child_P

    # def move_duplication_restart(self,pop):
    #     d_P = []
    #     d_PX = []
    #     for p in pop:
    #         if p.F_value not in d_PX:
    #             d_PX.append(p.F_value)
    #             p = self.idv.reset_one(p)
    #             d_P.append(p)
    #         else:
    #             # if method == 'restart':
    #             cp = self.re_start()
    #             # else:
    #             # p = self.idv.reset_one(p)
    #             # cp = self.multimode_mutation(p)
    #             d_P.append(cp)
    #     # while len(d_P) < self.popsize * 2:
    #     #         # ccp = self.re_start()
    #     #         # d_P.append(ccp)
    #     #         p = random.choice(pop)
    #     #         cp = self.idv.reset_one(p)
    #     #         ccp = self.multimode_mutation(cp)
    #     #         d_P.append(ccp)                

    #     return d_P

    def move_duplication(self,pop, pf):
        d_P = []
        d_PX = []
        for p in pop:
            if p.F_value not in d_PX:
                d_PX.append(p.F_value)
                p = self.idv.reset_one(p)
                d_P.append(p)

            # else:
            #     cp = self.idv.reset_one(p)
            #     ccp = self.multimode_mutation(cp)
            #     d_P.append(ccp)

        while len(d_P) < self.popsize * 2:
            if random.random() < 0.3 and len(pf) != 0:
                p = random.choice(pf)
                cp = self.idv.reset_one(p)
                ccp = self.multimode_mutation(cp)
                d_P.append(ccp)
            else:
                ccp = self.re_start()
                d_P.append(ccp)

        return d_P

        
    def re_start(self):
        p = self.idv.creat_one()
        solution = self.problem.initialsolution()
        p.X = solution[1]
        p.F_value = solution[0]
        return p
  

    def crossover(self, parent_a, parent_b):      
        feasible_p = []
        crossed = self.crossover_PMX(parent_a, parent_b)
        for p in crossed:
            # check feasible
            [flag1, infeasible_index, infeasible_j] = self.problem.check_feasible(p.X[0], p.X[1])
            if flag1 == False:
                [flag2, s_truck, s_drone] = self.problem.reassign(p.X[0], p.X[1], infeasible_j=infeasible_j, infeasible_index=infeasible_index)
                if flag2 == True:
                    p.X[0]    = s_truck
                    p.X[1]    = s_drone
                    p.F_value = self.problem.evaluate(p.X[0], p.X[1])
                    feasible_p.append(p)
            else:
                feasible_p.append(p)
        if feasible_p == []:           
            return [parent_a, parent_b]
        else:
            return feasible_p

       
    def crossover_PMX(self,p1, p2):
        # one-point crossover
        len_truck1 = len(p1.X[0])
        len_truck2 = len(p2.X[0])
        list1 = np.append(p1.X[0], p1.X[1][1])
        list2 = np.append(p2.X[0], p2.X[1][1])
        
        while True:
            k = random.randint(1, len(list1)-1)
            if list1[k] not in self.problem.truck_only_cus:
                if list2[k] not in self.problem.truck_only_cus:
                    break
        
        replace1 = copy.deepcopy(list1[k])
        replace2 = copy.deepcopy(list2[k])

        list1[k], list2[k] = replace2, replace1

        for i in range(list1.shape[0]):
            if i != k and list1[i] == replace2:
                list1[i] = replace1
        for i in range(list2.shape[0]):
            if i != k and list2[i] == replace1:
                list2[i] = replace2        

        cp1 = copy.deepcopy(p1)
        cp2 = copy.deepcopy(p2)
        cp1.X[0]    = list1[:len_truck1]
        cp1.X[1][1] = list1[len_truck1:]
        cp2.X[0]    = list2[:len_truck2]
        cp2.X[1][1] = list2[len_truck2:]

        return [cp1, cp2]