import os
import os.path
import sys
import numpy as np
import random
import json
import copy
import math
#from Mean_Vector_Util import Mean_vector

def distance_Manhattan(vector1,vector2):
    return np.sum(np.abs(vector1-vector2))

def distance_Euclidean(vector1,vector2):
    return np.sqrt(np.sum(np.square(vector1-vector2)))

def get_problemName(file_path):
    os.chdir(file_path)
    #print(os.path.abspath(os.curdir))
    all_file = os.listdir()
    return all_file

def read_data(problemName, file_path, wbl = 0, wbu = 0):
    data_dict = {}

    with open (file_path + problemName) as f:
        data = json.load(f)

    for key in data:
        data_dict[int(key)] = data[key]

    for nodeID in data_dict:
        eTW = data_dict[nodeID]['eTW']
        lTW = data_dict[nodeID]['lTW']
        earliest_time = eTW - wbl * (lTW - eTW)
        if earliest_time < 0:
            earliest_time = 0
        latest_time = lTW + wbu * (lTW - eTW)

        data_dict[nodeID]['earliest_time'] = earliest_time
        data_dict[nodeID]['latest_time'] = latest_time

    dataNum = len(data_dict)
    droneMatrix = np.zeros((dataNum, dataNum))
    truckMatrix = np.zeros((dataNum, dataNum))
    for i in range(dataNum):
        loci = np.array([data_dict[i]['latDeg'], data_dict[i]['lonDeg']])
        for j  in range(dataNum):
            locj = np.array([data_dict[j]['latDeg'],  data_dict[j]['lonDeg']])
            droneMatrix[i][j] = distance_Euclidean(loci,locj)
            truckMatrix[i][j] = distance_Manhattan(loci,locj)

    return [data_dict, droneMatrix, truckMatrix]

def nearest_neighbor(truck_cus,truck_time_matrix):
    truck_cus_tsp = []
   
    # depot
    cus_num = 0
    truck_cus_tsp.append(cus_num)
    
    # deal with other nodes
    while len(truck_cus_tsp) != len(truck_cus):
        shortest_time = float('inf')
        for cus in truck_cus:
            if cus not in truck_cus_tsp:
                to_node_time = truck_time_matrix[cus_num][cus]
                if shortest_time >= to_node_time:
                    shortest_time = to_node_time
                    temp_cus = cus
        truck_cus_tsp.append(temp_cus)
        cus_num = temp_cus

    return truck_cus_tsp

def nearest_neighbor_tw(truck_cus,truck_time_matrix, data_dict):
    truck_cus_tsp = []
   
    # depot
    cus_num = 0
    truck_cus_tsp.append(cus_num)
    t = 0
    
    # deal with other nodes
    while len(truck_cus_tsp) != len(truck_cus):
        shortest_time = float('inf')
        for cus in truck_cus:
            if cus not in truck_cus_tsp:
                earliest_tw = data_dict[cus]['earliest_time']
                latest_tw = data_dict[cus]['latest_time']
                to_node_time = truck_time_matrix[cus_num][cus]
                if shortest_time >= to_node_time and t >= earliest_tw and t <= latest_tw:
                    shortest_time = to_node_time
                    temp_cus = cus
        if shortest_time != float('inf'):
            truck_cus_tsp.append(temp_cus)
            cus_num = temp_cus
            t += shortest_time
        else:
            break

    while len(truck_cus_tsp) != len(truck_cus):
        shortest_time = float('inf')
        for cus in truck_cus:
            if cus not in truck_cus_tsp:
                earliest_tw = data_dict[cus]['earliest_time']
                latest_tw = data_dict[cus]['latest_time']
                to_node_time = truck_time_matrix[cus_num][cus]
                if shortest_time >= to_node_time:
                    shortest_time = to_node_time
                    temp_cus = cus

        truck_cus_tsp.append(temp_cus)
        cus_num = temp_cus

    return truck_cus_tsp

def is_dominate(a, b):
    # determine if a dominate b
    i = 0
    for av, bv in zip(a, b):
        if av < bv:
            i = i + 1
        if av > bv:
            return False
    if i != 0:
        return True
    return False

def is_weekly_dominate(a, b):
    # determine if a weekly dominate b
    i = 0
    for av, bv in zip(a, b):
        if av <= bv:
            i = i + 1
        if av > bv:
            return False
    if i != 0:
        return True
    return False

def evaluate_satisfaction(location_dict, node, reach_time):
    satisfaction = 0
    earliest_time = location_dict[int(node)]['earliest_time']
    latest_time = location_dict[int(node)]['latest_time']
    etw = location_dict[int(node)]['eTW']
    ltw = location_dict[int(node)]['lTW']
    if reach_time < earliest_time:
        satisfaction += 0
    elif earliest_time <= reach_time < etw:
        satisfaction += (reach_time - earliest_time)/(etw -earliest_time)
    elif etw <= reach_time <= ltw:
        satisfaction += 1
    elif ltw < reach_time <= latest_time:
        satisfaction += (latest_time - reach_time)/(latest_time -ltw)
    elif reach_time > latest_time:
        satisfaction += 0
    return satisfaction

#def shift(s_truck):
#    # Shif (1-0). A node is removed from its current position, and is inserted in a different position of a different tour.
#    s_truck = list(s_truck)
#    a_index = random.choice([i for i in range(len(s_truck))])
#    while True:
#        b_index = random.choice([i for i in range(len(s_truck))])
#        if a_index != b_index:
#            break
#    a_cus = s_truck[a_index]
#    s_truck.remove(a_cus)
#    s_truck.insert(b_index, a_cus)
#    s_truck = np.array(s_truck)
#    return s_truck

def relocation(truck_cus):
    a_index = random.choice([i for i in range(1,len(truck_cus))])
    while True:
        b_index = random.choice([i for i in range(1,len(truck_cus))])
        if a_index != b_index:
            break
    selected_cus = truck_cus[b_index]
    truck_cus.remove(selected_cus)
    truck_cus.insert(a_index, selected_cus)
    return truck_cus

def swap(truck_cus):
    # Swap (1-1). Two nodes exchange their positions.
    a_index = random.choice([i for i in range(1, len(truck_cus))])
    while True:
        b_index = random.choice([i for i in range(1, len(truck_cus))])
        if a_index != b_index:
            break
    truck_cus[a_index], truck_cus[b_index] = truck_cus[b_index], truck_cus[a_index]
    return truck_cus

def two_opt(truck_cus):
    # 2-Opt. Two non-adjacent arcs belonging to the same tour are removed, and two new arcs are added in order to find a new tour. The nodes invloved in the moves are selected randomly.
    a_index = random.choice([i for i in range(1, len(truck_cus))])
    while True:
        b_index = random.choice([i for i in range(1, len(truck_cus))])
        if a_index != b_index:
            break
    if a_index > b_index:
        path = truck_cus[b_index:a_index+1]
        start= b_index
        end = a_index
    else:
        path = truck_cus[a_index:b_index+1]
        start = a_index
        end = b_index
    rePath = path.copy()
    rePath = reversed(rePath)
    truck_cus[start:end+1] = rePath
    return truck_cus

def store_time(outpath, problemName, AlgorithmName, time):
    with open(outpath + AlgorithmName + '_' + 'Runtime.csv', 'a+') as f:
        time = str(time)
        data = [AlgorithmName, problemName, time]
        data = ','.join(data)
        f.write(data)
        f.write('\n')    

# def Load_W(h, Func_num, path = 'weights.csv'):
#     mv = Mean_vector(h, Func_num, path = path)
#     mv.generate()
#     W = np.loadtxt(fname=path)
#     return W

def normalization(value, max_value, min_value):
    return abs((value - min_value)/(max_value - min_value))

def generate_W(N, m = 2):
    w_matrix = np.zeros((N,m))
    for k in range(N):
        w_matrix[k,:] = np.array([(k+1-1)/(N-1), (N - k + 1)/(N-1)]) # k start from 1
    return w_matrix

def random_pick(some_list, probabilities): 
    x = random.uniform(0,1) 
    cumulative_probability = 0.0 
    for item, item_probability in zip(some_list, probabilities): 
         cumulative_probability += item_probability 
         if x < cumulative_probability:
               break 
    return item 